use crate::{
    db::{Connection, StatementContext, StatementRow, Transaction},
    schema::{
        datum::{Datum, QueryEquivalent, QueryEquivalentList},
        entity::{
            helpers::check_relation, Entity, EntityID, EntityPart, EntityPartList,
            EntityPartVisitor,
        },
        index::Index,
        relation::{LocalSide, RelationData},
        IDMap, Stored,
    },
    DBResult, Error,
};

use std::collections::HashMap;
use std::hash::{Hash, Hasher};

pub(crate) mod components;

pub(crate) fn insert<E: Entity>(conn: &Connection, value: &E) -> DBResult<E::ID> {
    struct InsertQuery<E: Entity>(std::marker::PhantomData<E>);

    conn.with_prepared(
        std::any::TypeId::of::<InsertQuery<E>>(),
        || {
            let mut part_names = String::new();
            let mut placeholders = String::new();
            struct PartNameVisitor<'a, E: Entity>(
                &'a mut String,
                &'a mut String,
                std::marker::PhantomData<E>,
            );
            impl<'a, E: Entity> EntityPartVisitor for PartNameVisitor<'a, E> {
                type Entity = E;
                fn visit<EP: EntityPart>(&mut self) {
                    // if this is a set-relation, then we don't actually want to do anything
                    // with it here; it doesn't have a column
                    if check_relation::<EP>() {
                        return;
                    }

                    if !self.0.is_empty() {
                        self.0.push_str(", ");
                        self.1.push_str(", ");
                    }
                    self.0.push('`');
                    self.0.push_str(EP::part_name());
                    self.0.push('`');
                    self.1.push('?');
                }
            }

            E::accept_part_visitor(&mut PartNameVisitor(
                &mut part_names,
                &mut placeholders,
                Default::default(),
            ));

            format!(
                "INSERT INTO `{}` ({}) VALUES ({}) RETURNING `id`",
                E::entity_name(),
                part_names,
                placeholders
            )
        },
        |mut ctx| {
            struct PartBinder<'a, 'b, E: Entity>(
                &'a mut StatementContext<'b>,
                i32,
                std::marker::PhantomData<E>,
            );
            impl<'a, 'b, E: Entity> EntityPartVisitor for PartBinder<'a, 'b, E> {
                type Entity = E;
                fn visit_datum<EP: EntityPart>(&mut self, datum: &EP::Datum) {
                    // skip relations, as with the query preparation above
                    if check_relation::<EP>() {
                        return;
                    }

                    datum.bind_to(self.0, self.1);
                    self.1 += 1;
                }
            }

            value.accept_part_visitor_ref(&mut PartBinder(&mut ctx, 1, Default::default()));

            ctx.run()?
                .ok_or(Error::InternalError("No result row from INSERT query"))
                .map(|r| <E::ID>::from_raw(r.read(0).expect("couldn't read resulting ID")))
        },
    )
}

pub(crate) fn insert_and_return<E: Entity>(conn: &Connection, mut value: E) -> DBResult<Stored<E>> {
    let id = insert(conn, &value)?;

    // update relation data in all fields
    struct DatumWalker<'l, E: Entity>(&'l Connection, i64, std::marker::PhantomData<E>);
    impl<'l, E: Entity> EntityPartVisitor for DatumWalker<'l, E> {
        type Entity = E;
        fn visit_datum_mut<EP: EntityPart>(&mut self, datum: &mut EP::Datum) {
            datum.update_adata(RelationData {
                conn: self.0.clone(),
                part_name: EP::part_name(),
                local_name: <EP::Entity as Entity>::entity_name(),
                local_id: self.1,
            });
        }
    }

    value.accept_part_visitor_mut(&mut DatumWalker(conn, id.into_raw(), Default::default()));

    Ok(Stored::new(conn.clone(), id, value))
}

pub(crate) fn update_entity<E: Entity>(conn: &Connection, value: &Stored<E>) -> DBResult<()> {
    struct UpdateQuery<E: Entity>(std::marker::PhantomData<E>);

    conn.with_prepared(
        std::any::TypeId::of::<UpdateQuery<E>>(),
        || {
            let mut set_columns = String::new();
            struct PartNameVisitor<'a, E: Entity>(&'a mut String, std::marker::PhantomData<E>);
            impl<'a, E: Entity> EntityPartVisitor for PartNameVisitor<'a, E> {
                type Entity = E;
                fn visit<EP: EntityPart>(&mut self) {
                    // if this is a set-relation, then we don't actually want to do anything
                    // with it here; it doesn't have a column
                    if check_relation::<EP>() {
                        return;
                    }

                    if !self.0.is_empty() {
                        self.0.push_str(", ");
                    }
                    self.0.push('`');
                    self.0.push_str(EP::part_name());
                    self.0.push_str("` = ?");
                }
            }

            E::accept_part_visitor(&mut PartNameVisitor(&mut set_columns, Default::default()));
            format!(
                "UPDATE `{entity_name}` SET {set_columns} WHERE `id` = ?",
                entity_name = E::entity_name()
            )
        },
        |mut ctx| {
            struct PartBinder<'a, 'b, E: Entity>(
                &'a mut StatementContext<'b>,
                &'a mut i32,
                std::marker::PhantomData<E>,
            );
            impl<'a, 'b, E: Entity> EntityPartVisitor for PartBinder<'a, 'b, E> {
                type Entity = E;
                fn visit_datum<EP: EntityPart>(&mut self, datum: &EP::Datum) {
                    // skip relations, as with the query preparation above
                    if check_relation::<EP>() {
                        return;
                    }

                    datum.bind_to(self.0, *self.1);
                    *self.1 += 1;
                }
            }

            // first bind all the updating clauses
            let mut index = 1;
            value.accept_part_visitor_ref(&mut PartBinder(
                &mut ctx,
                &mut index,
                Default::default(),
            ));

            // then bind the id
            value.id().bind_to(&mut ctx, index);

            ctx.run()?;

            Ok(())
        },
    )
}

#[derive(Debug, Clone, Copy, Hash, PartialEq, Eq)]
pub(crate) enum QueryPart {
    Root,
    Columns,
    From,
    Set,
    Join,
    Where,
    Trailing,
}

#[derive(Debug)]
pub struct Query {
    parts: HashMap<QueryPart, Vec<String>>,
}

impl Query {
    pub(crate) fn new() -> Self {
        Self {
            parts: Default::default(),
        }
    }

    pub(crate) fn attach(mut self, qp: QueryPart, val: String) -> Self {
        self.attach_mut(qp, val);
        self
    }

    pub(crate) fn replace(mut self, qp: QueryPart, val: String) -> Self {
        self.parts.remove(&qp);
        self.attach(qp, val)
    }

    pub(crate) fn attach_mut(&mut self, qp: QueryPart, val: String) {
        self.parts.entry(qp).or_default().push(val);
    }

    pub(crate) fn assemble(mut self) -> String {
        let root = self.parts.remove(&QueryPart::Root).unwrap().remove(0);

        let columns_ = match self.parts.remove(&QueryPart::Columns) {
            None => String::new(),
            Some(v) => v
                .into_iter()
                .reduce(|a, b| format!("{}, {}", a, b))
                .unwrap(),
        };

        let from_ = match self.parts.remove(&QueryPart::From) {
            None => String::new(),
            Some(v) => {
                format!(
                    "FROM {}",
                    v.into_iter()
                        .reduce(|a, b| format!("{}, {}", a, b))
                        .unwrap()
                )
            }
        };

        let set_ = match self.parts.remove(&QueryPart::Set) {
            None => String::new(),
            Some(v) => {
                format!(
                    "SET {}",
                    v.into_iter()
                        .reduce(|a, b| format!("{}, {}", a, b))
                        .unwrap()
                )
            }
        };

        let join_ = match self.parts.remove(&QueryPart::Join) {
            None => String::new(),
            Some(v) => v
                .into_iter()
                .map(|j| format!("INNER JOIN {}", j))
                .reduce(|a, b| format!("{} {}", a, b))
                .unwrap(),
        };

        let where_ = match self.parts.remove(&QueryPart::Where) {
            None => String::new(),
            Some(v) => {
                format!(
                    "WHERE {}",
                    v.into_iter()
                        .reduce(|a, b| format!("{} AND {}", a, b))
                        .unwrap()
                )
            }
        };

        let trailing_ = match self.parts.remove(&QueryPart::Trailing) {
            None => String::new(),
            Some(v) => v.into_iter().reduce(|a, b| format!("{} {}", a, b)).unwrap(),
        };

        format!(
            "{} {} {} {} {} {} {}",
            root, columns_, from_, set_, join_, where_, trailing_
        )
    }
}

pub(crate) struct RelationNames {
    local_name: &'static str,
    remote_name: &'static str,
    part_name: &'static str,
    dist_name: &'static str,
    domain_name: &'static str,
    range_name: &'static str,
    local_field: &'static str,
    remote_field: &'static str,
}

impl RelationNames {
    fn collect<AI: RelationInterface>(iface: &AI) -> DBResult<RelationNames> {
        let rdata = iface.get_data()?;
        let local_name = rdata.local_name;
        let remote_name = <AI::RemoteEntity>::entity_name();
        let part_name = rdata.part_name;
        let dist_name = iface.get_distinguishing_name()?;

        let (domain_name, range_name) = match AI::SIDE {
            LocalSide::Domain => (local_name, remote_name),
            LocalSide::Range => (remote_name, local_name),
        };
        let (local_field, remote_field) = match AI::SIDE {
            LocalSide::Domain => ("domain", "range"),
            LocalSide::Range => ("range", "domain"),
        };
        Ok(Self {
            local_name,
            remote_name,
            part_name,
            dist_name,
            domain_name,
            range_name,
            local_field,
            remote_field,
        })
    }

    fn relation_name(&self) -> String {
        format!(
            "{domain_name}_{range_name}_relation_{dist_name}",
            domain_name = self.domain_name,
            range_name = self.range_name,
            dist_name = self.dist_name
        )
    }
}

fn hash_of<T: Hash>(val: T) -> u64 {
    let mut hasher = std::collections::hash_map::DefaultHasher::new();
    val.hash(&mut hasher);
    hasher.finish()
}

fn do_connect<Remote: Entity>(
    rdata: &RelationData,
    an: RelationNames,
    remote_id: Remote::ID,
) -> DBResult<()> {
    rdata.conn.with_prepared(
        hash_of(("connect", an.local_name, an.remote_name, an.part_name)),
        || {
            format!(
                "insert into `{relation_name}` (`{local_field}`, `{remote_field}`) values (?, ?) returning (`id`)",
                relation_name = an.relation_name(),
                local_field = an.local_field,
                remote_field = an.remote_field
            )
        },
        |ctx| {
            ctx.bind(1, rdata.local_id)?;
            ctx.bind(2, remote_id.into_raw())?;

            ctx.run()?
                .ok_or(Error::ConstraintViolation("Relation entry uniqueness".to_string()))
                .map(|_| ())
        },
    )
}

/// Relation map generic interface trait.
pub trait RelationInterface: 'static {
    /// The type of the entity on the non-local end of the relation.
    type RemoteEntity: Entity;

    #[doc(hidden)]
    fn get_data(&self) -> DBResult<&RelationData>;
    #[doc(hidden)]
    fn get_distinguishing_name(&self) -> DBResult<&'static str>;

    /// Which side is the "local" side of the relation.
    const SIDE: LocalSide;

    /// Query this entity type without the relation filter.
    fn query_all(&self) -> impl Queryable<EntityOutput = Self::RemoteEntity> {
        components::TableComponent::<Self::RemoteEntity>::new(self.get_data().unwrap().conn.clone())
    }

    /// Attempt to connect the contextual instance to a remote instance.
    fn connect_to(&self, remote_id: <Self::RemoteEntity as Entity>::ID) -> DBResult<()>
    where
        Self: Sized,
    {
        let rdata = self.get_data()?;
        let an = RelationNames::collect::<Self>(self)?;

        let txn = Transaction::new(&rdata.conn)?;

        do_connect::<Self::RemoteEntity>(rdata, an, remote_id)?;

        txn.commit()
    }

    /// Attempt to disconnect the contextual instance from a remote instance.
    fn disconnect_from(&self, remote_id: <Self::RemoteEntity as Entity>::ID) -> DBResult<()>
    where
        Self: Sized,
    {
        let rdata = self.get_data()?;
        let an = RelationNames::collect::<Self>(self)?;

        let txn = Transaction::new(&rdata.conn)?;

        // second, add to the relation table
        rdata.conn.with_prepared(
            hash_of(("disconnect", an.local_name, an.remote_name, an.part_name)),
            || {
                format!(
                    "delete from `{relation_name}` where `{local_field}` = ? and `{remote_field}` = ?",
                    relation_name = an.relation_name(),
                    local_field = an.local_field,
                    remote_field = an.remote_field
                )
            },
            |ctx| {
                ctx.bind(1, rdata.local_id)?;
                ctx.bind(2, remote_id.into_raw())?;

                ctx.run().map(|_| ())
            },
        )?;

        txn.commit()
    }
}

// ----------------------------------------------------------------------
// New query interface
// ----------------------------------------------------------------------

/// Represents a context in which we can insert an entity type `E`.
pub trait Insertable<E: Entity> {
    /// Insert an entity instance and return its new ID.
    fn insert(&self, value: E) -> DBResult<E::ID>;
    /// Insert an entity instance and return a [`Stored`] instance that can be used to synchronize
    /// its values back into the database later.
    fn insert_and_return(&self, value: E) -> DBResult<Stored<E>>;
}

impl<AI: RelationInterface> Insertable<AI::RemoteEntity> for AI {
    fn insert(&self, value: AI::RemoteEntity) -> DBResult<<AI::RemoteEntity as Entity>::ID>
    where
        Self: Sized,
    {
        // we're doing two things:
        // - inserting the entity into the target table
        // - adding the relation row into the relation table

        let rdata = self.get_data()?;
        let an = RelationNames::collect::<Self>(self)?;

        let txn = Transaction::new(&rdata.conn)?;

        // so first, into the remote table
        let remote_id = insert(&rdata.conn, &value)?;
        // then the relation
        do_connect::<AI::RemoteEntity>(rdata, an, remote_id)?;

        txn.commit()?;

        Ok(remote_id)
    }

    fn insert_and_return(&self, value: AI::RemoteEntity) -> DBResult<Stored<AI::RemoteEntity>>
    where
        Self: Sized,
    {
        // we're doing two things:
        // - inserting the entity into the target table
        // - adding the relation row into the relation table

        let rdata = self.get_data()?;
        let an = RelationNames::collect::<Self>(self)?;

        let txn = Transaction::new(&rdata.conn)?;

        // so first, into the remote table
        let remote = insert_and_return(&rdata.conn, value)?;
        // then the relation
        do_connect::<AI::RemoteEntity>(rdata, an, remote.id())?;

        txn.commit()?;

        Ok(remote)
    }
}

pub trait IDContainer<T: Entity>: 'static + IntoIterator<Item = T::ID> {
    fn assemble_from(ctx: StatementContext<'_>) -> DBResult<Self>
    where
        Self: Sized;
}

pub trait OutputContainer<T: Entity>: 'static + IntoIterator<Item = Stored<T>> {
    type IDContainer: IDContainer<T>;
    type ReplacedEntity<N: Entity>: OutputContainer<N>;
    fn assemble_from(conn: &Connection, stmt: StatementContext<'_>) -> DBResult<Self>
    where
        Self: Sized;
}

fn assemble_id<T: Entity>(row: StatementRow) -> T::ID {
    <T::ID>::from_raw(row.read::<i64>(0).expect("couldn't read ID"))
}

fn assemble_single<T: Entity>(conn: &Connection, row: &mut StatementRow) -> Stored<T> {
    let id = row.read::<i64>(0).expect("couldn't read ID");
    let datum_list = <T::Parts>::build_datum_list(conn, row).expect("couldn't build datum list");
    Stored::new(conn.clone(), T::ID::from_raw(id), T::build(datum_list))
}

impl<T: Entity> IDContainer<T> for Option<T::ID> {
    fn assemble_from(ctx: StatementContext<'_>) -> DBResult<Self>
    where
        Self: Sized,
    {
        Ok(ctx.run()?.map(assemble_id::<T>))
    }
}

impl<T: Entity> OutputContainer<T> for Option<Stored<T>> {
    type IDContainer = Option<T::ID>;
    type ReplacedEntity<N: Entity> = Option<Stored<N>>;

    fn assemble_from(conn: &Connection, ctx: StatementContext<'_>) -> DBResult<Self>
    where
        Self: Sized,
    {
        Ok(ctx.run()?.map(|mut r| assemble_single(conn, &mut r)))
    }
}

impl<T: Entity> IDContainer<T> for Vec<T::ID> {
    fn assemble_from(ctx: StatementContext<'_>) -> DBResult<Self>
    where
        Self: Sized,
    {
        ctx.iter()
            .map(|r| r.map(assemble_id::<T>))
            .collect::<Result<Vec<_>, Error>>()
    }
}

impl<T: Entity> OutputContainer<T> for Vec<Stored<T>> {
    type IDContainer = Vec<T::ID>;
    type ReplacedEntity<N: Entity> = Vec<Stored<N>>;

    fn assemble_from(conn: &Connection, ctx: StatementContext<'_>) -> DBResult<Self>
    where
        Self: Sized,
    {
        ctx.iter()
            .map(|r| r.map(|mut s| assemble_single(conn, &mut s)))
            .collect::<Result<Vec<_>, Error>>()
    }
}

/// Represents a searchable context of a given entity.
pub trait Queryable: Clone {
    /// The entity that results from a search in this context.
    type EntityOutput: Entity;
    /// How results will be provided. This is either a `Vec` or an `Option`.
    type OutputContainer: OutputContainer<Self::EntityOutput>;
    /// A `'static`-version of `Self`, used for `TypeId`-based caching.
    type StaticVersion: Queryable + 'static;

    /// Construct a concrete SQL query from this abstract type representation.
    #[doc(hidden)]
    fn build(&self) -> Query;
    /// Bind into any required placeholders to 'fill' an instance created by [`build`].
    #[doc(hidden)]
    fn bind(&self, stmt: &mut StatementContext, index: &mut i32);
    #[doc(hidden)]
    fn conn(&self) -> &Connection;

    // ----------------------------------------------------------------------
    // Verbs
    // ----------------------------------------------------------------------
    /// Count all entities in the current context.
    ///
    /// Returns the number of entities.
    fn count(self) -> DBResult<usize>
    where
        Self: Sized,
    {
        let txn = Transaction::new(self.conn())?;
        struct CountTag;
        let out = self.conn().with_prepared(
            std::any::TypeId::of::<(Self::StaticVersion, CountTag)>(),
            || {
                self.build()
                    .replace(
                        QueryPart::Columns,
                        format!(
                            "COUNT(DISTINCT `{}`.`id`)",
                            Self::EntityOutput::entity_name()
                        ),
                    )
                    .assemble()
            },
            |mut ctx| {
                // starting index is 1
                let mut index = 1;
                self.bind(&mut ctx, &mut index);

                Ok(ctx
                    .run()?
                    .ok_or(Error::InternalError("no resulting rows from COUNT query"))?
                    .read::<i64>(0)? as usize)
            },
        )?;
        txn.commit()?;
        Ok(out)
    }
    /// Get all entities in the current context.
    fn get(self) -> DBResult<Self::OutputContainer>
    where
        Self: Sized,
    {
        let txn = Transaction::new(self.conn())?;
        struct GetTag;
        let out = self.conn().with_prepared(
            std::any::TypeId::of::<(Self::StaticVersion, GetTag)>(),
            || self.build().assemble(),
            |mut ctx| {
                // starting index is 1
                let mut index = 1;
                self.bind(&mut ctx, &mut index);

                <Self::OutputContainer>::assemble_from(self.conn(), ctx)
            },
        )?;
        txn.commit()?;
        Ok(out)
    }
    /// Get IDs of all entities in the current context.
    fn get_ids(
        self,
    ) -> DBResult<<Self::OutputContainer as OutputContainer<Self::EntityOutput>>::IDContainer>
    where
        Self: Sized,
    {
        let txn = Transaction::new(self.conn())?;
        struct GetIDTag;
        let out = self.conn().with_prepared(
            std::any::TypeId::of::<(Self::StaticVersion, GetIDTag)>(),
            || {
                self.build()
                    .replace(
                        QueryPart::Columns,
                        format!(
                            "`{}`.`id`",
                            Self::EntityOutput::entity_name()
                        ),
                    )
                    .assemble()
            },
            |mut ctx| {
                // starting index is 1
                let mut index = 1;
                self.bind(&mut ctx, &mut index);

                <<Self::OutputContainer as OutputContainer<Self::EntityOutput>>::IDContainer>::assemble_from(ctx)
            },
        )?;
        txn.commit()?;
        Ok(out)
    }
    /// Delete all entities in the current context.
    fn delete(self) -> DBResult<()>
    where
        Self: Sized,
    {
        let txn = Transaction::new(self.conn())?;
        struct DeleteTag;
        self.conn().with_prepared(
            std::any::TypeId::of::<(Self::StaticVersion, DeleteTag)>(),
            || {
                format!(
                    "DELETE FROM `{}` WHERE `id` = ({})",
                    Self::EntityOutput::entity_name(),
                    self.build()
                        .replace(
                            QueryPart::Columns,
                            format!("`{}`.`id`", Self::EntityOutput::entity_name())
                        )
                        .assemble()
                )
            },
            |mut ctx| {
                // starting index is 1
                let mut index = 1;
                self.bind(&mut ctx, &mut index);

                ctx.run()?;
                Ok(())
            },
        )?;
        txn.commit()
    }

    // ----------------------------------------------------------------------
    // Filtering methods
    // ----------------------------------------------------------------------
    /// Filter using the keying index on the entity.
    fn keyed(
        self,
        values: impl QueryEquivalentList<
            <<Self::EntityOutput as Entity>::Keys as EntityPartList>::DatumList,
        >,
    ) -> impl Queryable<
        EntityOutput = Self::EntityOutput,
        OutputContainer = Option<Stored<Self::EntityOutput>>,
    >
    where
        Self: Sized,
    {
        components::IndexComponent::<_, _, <Self::EntityOutput as Entity>::Keys, _>::new(
            self, values,
        )
    }

    /// Filter using an arbitrary index on the entity.
    fn indexed<EPL: EntityPartList<Entity = Self::EntityOutput>>(
        self,
        _index: &Index<Self::EntityOutput, EPL>,
        values: impl QueryEquivalentList<EPL::DatumList>,
    ) -> impl Queryable<
        EntityOutput = Self::EntityOutput,
        OutputContainer = Option<Stored<Self::EntityOutput>>,
    >
    where
        Self: Sized,
    {
        components::IndexComponent::<_, _, EPL, _>::new(self, values)
    }

    /// Filter using an arbitrary column on the entity.
    fn with<EP: EntityPart<Entity = Self::EntityOutput>>(
        self,
        part: EP,
        value: impl QueryEquivalent<EP::Datum>,
    ) -> impl Queryable<EntityOutput = Self::EntityOutput, OutputContainer = Self::OutputContainer>
    where
        Self: Sized,
    {
        components::WithComponent::new(self, part, value)
    }

    /// Ask to return at most a single result
    fn first(
        self,
    ) -> impl Queryable<
        EntityOutput = Self::EntityOutput,
        OutputContainer = Option<Stored<Self::EntityOutput>>,
    >
    where
        Self: Sized,
    {
        components::SingleComponent::new(self)
    }

    // ----------------------------------------------------------------------
    // Relation-following and joining methods
    // ----------------------------------------------------------------------
    /// Join based on an existing relation
    fn join<
        AD: RelationInterface + Datum,
        EP: EntityPart<Entity = Self::EntityOutput, Datum = AD>,
    >(
        self,
        part: EP,
    ) -> impl Queryable<EntityOutput = AD::RemoteEntity, OutputContainer = Vec<Stored<AD::RemoteEntity>>>
    where
        Self: Sized,
    {
        components::JoinComponent::<AD::RemoteEntity, Self::EntityOutput, _, Self>::new(self, part)
    }

    /// Follow a foreign key.
    fn foreign<EP: EntityPart<Entity = Self::EntityOutput>>(
        self,
        part: EP,
    ) -> impl Queryable<
        EntityOutput = <EP::Datum as EntityID>::Entity,
        OutputContainer = <Self::OutputContainer as OutputContainer<Self::EntityOutput>>::ReplacedEntity<<EP::Datum as EntityID>::Entity>,
    >
    where
        Self: Sized,
        EP::Datum: EntityID,
    {
        components::ForeignComponent::<_, EP, Self>::new(self, part)
    }
}

// Generic implementations for all IDMaps
impl<'a, T: Entity> Queryable for &'a IDMap<T> {
    type EntityOutput = T;
    type OutputContainer = Vec<Stored<T>>;
    type StaticVersion = &'static IDMap<T>;

    fn build(&self) -> Query {
        Query::new()
            .attach(QueryPart::Root, "SELECT DISTINCT".into())
            .attach(QueryPart::Columns, "*".into())
            .attach(QueryPart::From, format!("`{}`", T::entity_name()))
    }
    fn bind(&self, _stmt: &mut StatementContext, _index: &mut i32) {}

    fn conn(&self) -> &Connection {
        &self.conn
    }
}

// Generic implementation for all relation specification types
impl<'a, AI: RelationInterface> Queryable for &'a AI {
    type EntityOutput = AI::RemoteEntity;
    type OutputContainer = Vec<Stored<AI::RemoteEntity>>;
    type StaticVersion = &'static AI;

    fn build(&self) -> Query {
        let anames = RelationNames::collect(*self).unwrap();
        let relation_name = anames.relation_name();
        Query::new()
            .attach(QueryPart::Root, "SELECT DISTINCT".into())
            .attach(QueryPart::Columns, format!("`{}`.*", anames.remote_name))
            .attach(QueryPart::From, format!("`{}`", relation_name))
            .attach(
                QueryPart::Join,
                format!(
                    "`{}` ON `{}`.`id` = `{}`.`{}`",
                    anames.remote_name, anames.remote_name, relation_name, anames.remote_field
                ),
            )
            .attach(
                QueryPart::Where,
                format!("`{}`.`{}` = ?", relation_name, anames.local_field),
            )
    }
    fn bind(&self, ctx: &mut StatementContext, index: &mut i32) {
        let rdata = self
            .get_data()
            .expect("binding query for relation with no data");

        ctx.bind(*index, rdata.local_id)
            .expect("couldn't bind relation id");
        *index += 1;
    }

    fn conn(&self) -> &Connection {
        &self.get_data().unwrap().conn
    }
}

impl<E: Entity, EPL: EntityPartList<Entity = E>> Index<E, EPL> {
    /// Perform a search through this index
    ///
    /// Note that this is simply sugar for `idmap_instance.indexed(self, values)`.
    pub fn search<'a>(
        &'a self,
        values: impl 'a + QueryEquivalentList<EPL::DatumList>,
    ) -> impl 'a + Queryable<EntityOutput = E, OutputContainer = Option<Stored<E>>> {
        self.indexed(self, values)
    }
}

impl<'a, E: Entity, EPL: EntityPartList<Entity = E>> Queryable for &'a Index<E, EPL> {
    type EntityOutput = E;
    type OutputContainer = Vec<Stored<E>>;
    type StaticVersion = &'static Index<E, EPL>;

    fn build(&self) -> Query {
        Query::new()
            .attach(QueryPart::Root, "SELECT DISTINCT".into())
            .attach(QueryPart::Columns, "*".into())
            .attach(QueryPart::From, format!("`{}`", E::entity_name()))
    }
    fn bind(&self, _stmt: &mut StatementContext, _index: &mut i32) {}

    fn conn(&self) -> &Connection {
        &self.conn
    }
}

#[cfg(test)]
mod query_build_test {
    #[test]
    fn simple_construction() {}
}
